/**********************************************************************
 *
 *  G_quant_init (quant)
 *       struct Quant *quant;
 *
 *  initializes new quantization structure. calls
 *  G_quant_clear() before it returns.
 *  
 *  note: dies if G_malloc dies.
 *
 **********************************************************************
 *
 *  void
 *  G_quant_free (q)
 *  
 *       struct Quant *q;
 *
 *  resets the number of defined rules to 0 and free's space allocated
 *  for rules. calls G_quant_clear ().
 *
 **********************************************************************
 *
 *  void
 *  G_quant_clear (q)
 *  
 *       struct Quant *q;
 *
 *  resets the number of defined rules and number of infinite rules to 0. 
 *
 **********************************************************************
 *
 *  int
 *  G_quant_get_limits (q, dMin, dMax, cMin, cMax)
 *  
 *       struct Quant *q;
 *       DCELL *dMin, *dMax;
 *       CELL *cMin, *cMax;
 *
 *  returns the minimum and maximum cell and dcell values of all
 *  the ranges defined.
 *  
 *  returns: -1 if after G_quant_init (), or any call to 
 *                 G_quant_clear () or G_quant_free () 
 *                 no explicit rules have been added
 *                 In this case the returned minimum and maximum 
 *                 CELL and DCELL values are undefined.
 *            1 otherwise. in this case the values returned correspond
 *                 to the extreme values of the defined rules 
 *
 **********************************************************************
 *  
 *  int
 *  G_quant_nof_rules (q)
 *  
 *       struct Quant *q;
 *  
 *  returns the number of quantization rules defined. This number does
 *  not include the 2 infinite intervals.
 *  
 **********************************************************************
 *  
 *  void
 *  G_quant_get_ith_rule (q, i, dLow, dHigh, cLow, cHigh)
 *  
 *       struct Quant *q;
 *       int i;
 *       DCELL *dLow, *dHigh;
 *       CELL *cLow, *cHigh;
 *  
 *  returns the i'th quantization rule, for 0 <= i < G_quant_nof_rules().
 *  a larger value for i means that the rule has been added later.
 *  
 **********************************************************************
 *   void
 *   G_quant_set_neg_infinite_rule (q, dLeft, c)
 *
 *       struct Quant *q;
 *       DCELL dLeft;
 *       CELL c;
 *
 *   defines a rule for values "dLeft" and smaller. values in this range
 *   are mapped to "c" if none of the "finite" quantization rules applies.
 *
 * **********************************************************************
 *
 *  int
 *  G_quant_get_neg_infinite_rule (q, dLeft, c)
 *
 *       struct Quant *q;
 *       DCELL *dLeft;
 *       CELL *c;
 *
 *  returns in "dLeft" and "c" the rule values for the negative infinite
 *  interval (see G_quant_set_neg_infinite_rule ()).
 *
 *  returns: 0 if this rule is not defined
 *           1 otherwise.
 *
 **********************************************************************
 *   void
 *   G_quant_set_pos_infinite_rule (q, dRight, c)
 *
 *       struct Quant *q;
 *       DCELL dRight;
 *       CELL c;
 *
 *   defines a rule for values "dRight" and larger. values in this range
 *   are mapped to "c" if none of the "finite" quantization rules or the
 *   negative infinite rule applies.
 *
 * **********************************************************************
 *
 *  int
 *  G_quant_get_pos_infinite_rule (q, dRight, c)
 *
 *       struct Quant *q;
 *       DCELL *dRight;
 *       CELL *c;
 *
 *  returns in "dRight" and "c" the rule values for the positive infinite
 *  interval (see G_quant_set_pos_infinite_rule ()).
 *
 *  returns: 0 if this rule is not defined
 *           1 otherwise.
 *
 **********************************************************************
 *  
 *  void
 *  G_quant_reverse_rule_order (q)
 *
 *        struct Quant *q;
 *
 *  reverses the order in which the qunatization rules are stored. (see
 *  also G_quant_get_ith_rule () and G_quant_perform_d ()).
 *  
 **********************************************************************
 *  
 *  void
 *  G_quant_add_rule (q, dLow, dHigh, cLow, cHigh)
 *  
 *       struct Quant *q;
 *       DCELL dLow, dHigh;
 *       CELL cLow, cHigh;
 *  
 *  adds a new rule to the set of quantization rules. if dLow < dHigh
 *  the rule will be stored with the low and high values interchanged.
 *  
 *  Note: currently no cleanup of rules is performed, i.e. redundant
 *        rules are not removed.
 *  
 **********************************************************************
 *  
 *  CELL
 *  G_quant_get_cell_value (q, cellValue)
 *  
 *       struct Quant *q;
 *       DCELL *cellValue;
 *  
 *  returns in "cell" the quantized CELL values corresponding to the
 *  DCELL value "cellValue".
 *
 *  if several quantization rules apply for cellValue, the one which has 
 *  been inserted latest (i.e. the one of them which is returned by 
 *  G_quant_get_ith_rule() for the largest i) is used. if no such rule
 *  applies the cellValue is first tested against the negative infinite
 *  rule, and finally against the positive infinite rule. if none of
 *  these rules apply, NO_DATA is returned. the actual value of NO_DATA 
 *  is found by calling G_c_set_null_value().
 *  
 *  if after G_quant_init (), or any call to G_quant_clear () or 
 *  G_quant_free () neither G_quant_add_rule (),
 *  G_quant_set_neg_infinite_rule (),  G_quant_set_pos_infinite_rule ()
 *  are used NO_DATA is returned independently 
 *  of cellValue.
 *
 **********************************************************************
 *  
 *  void
 *  G_quant_perform_d (q, dcell, cell, n)
 *  
 *       struct Quant *q;
 *       DCELL *dcell;
 *       CELL *cell;
 *       int n;
 *  
 *  returns in "cell" the quantized CELL values corresponding to the
 *  DCELL values stored in "dcell". the number of elements quantized
 *  is n. quantization is performed by repeated application of 
 *  G_quant_get_cell_value ().
 *  
 **********************************************************************
 *  
 *  void
 *  G_quant_perform_f (q, fcell, cell, n)
 *  
 *       struct Quant *q;
 *       FCELL *fcell;
 *       CELL *cell;
 *       int n;
 *  
 *  same as G_quant_perform_d (), except the type.
 *  
 **********************************************************************/

/*--------------------------------------------------------------------------*/

/*
 the quantization table is stored as a linear array. rules are added starting
 from index 0. redundant rules are not eliminated. rules are tested from the 
 highest index downto 0. there are two "infinite" rules. support is provided 
 to reverse the order of the rules.
*/

/*--------------------------------------------------------------------------*/

#include "gis.h"

/*--------------------------------------------------------------------------*/

#define MIN(a,b) (a < b ? a : b)
#define MAX(a,b) (a > b ? a : b)
#define NO_DATA (G_set_c_null_value (&tmp, 1), (CELL) tmp)

#define NO_LEFT_INFINITE_RULE (! q->infiniteLeftSet)
#define NO_RIGHT_INFINITE_RULE (! q->infiniteRightSet)
#define NO_FINITE_RULE (q->nofRules <= 0)
#define NO_EXPLICIT_RULE (NO_FINITE_RULE && \
			  NO_LEFT_INFINITE_RULE && NO_RIGHT_INFINITE_RULE)

/*--------------------------------------------------------------------------*/

void
G_quant_clear (q)

     struct Quant *q;

{
  q->nofRules = 0;
  q->infiniteRightSet = q->infiniteLeftSet = 0;
}

/*--------------------------------------------------------------------------*/

void
G_quant_free (q)

     struct Quant *q;

{
  G_quant_clear (q);

  if (q->maxNofRules > 0) free (q->table);

  q->maxNofRules = 0;
}

/*--------------------------------------------------------------------------*/

G_quant_init (quant)
  struct Quant *quant;
{
  quant->maxNofRules = 0;
  G_quant_clear (quant);
  
  return 1;
}

/*--------------------------------------------------------------------------*/

static void
quant_set_limits (q, dLow, dHigh, cLow, cHigh)

     struct Quant *q;
     DCELL dLow, dHigh;
     CELL cLow, cHigh;

{
  q->dMin = dLow;
  q->dMax = dHigh;
  q->cMin = cLow;
  q->cMax = cHigh;
}

/*--------------------------------------------------------------------------*/

static void
quant_update_limits (q, dLow, dHigh, cLow, cHigh)

     struct Quant *q;
     DCELL dLow, dHigh;
     CELL cLow, cHigh;

{
  if (NO_EXPLICIT_RULE) {
    quant_set_limits (q, dLow, dHigh, cLow, cHigh);
    return;
  }

  q->dMin = MIN (q->dMin, MIN (dLow, dHigh));
  q->dMax = MAX (q->dMax, MAX (dLow, dHigh));
  q->cMin = MIN (q->cMin, MIN (cLow, cHigh));
  q->cMax = MAX (q->cMax, MAX (cLow, cHigh));
}

/*--------------------------------------------------------------------------*/

int
G_quant_get_limits (q, dMin, dMax, cMin, cMax)

     struct Quant *q;
     DCELL *dMin, *dMax;
     CELL *cMin, *cMax;

{
  if (NO_EXPLICIT_RULE) return -1;

  *dMin = q->dMin; *dMax = q->dMax;
  *cMin = q->cMin; *cMax = q->cMax;

  return 1;
}

/*--------------------------------------------------------------------------*/

int
G_quant_nof_rules (q)

     struct Quant *q;

{
  return q->nofRules;
}

/*--------------------------------------------------------------------------*/

void
G_quant_get_ith_rule (q, i, dLow, dHigh, cLow, cHigh)

     struct Quant *q;
     int i;
     DCELL *dLow, *dHigh;
     CELL *cLow, *cHigh;

{
  *dLow = q->table[i].dLow;
  *dHigh = q->table[i].dHigh;
  *cLow = q->table[i].cLow;
  *cHigh = q->table[i].cHigh;
}

/*--------------------------------------------------------------------------*/

static void
quant_table_increase (q)

     struct Quant *q;

{
  if (q->nofRules < q->maxNofRules) return;

  if (q->maxNofRules == 0) {
    q->maxNofRules = 50;
    q->table = (struct Quant_table *) 
                 G_malloc (q->maxNofRules * sizeof (struct Quant_table));
  } else {
    q->maxNofRules += 50;
    q->table = (struct Quant_table *) 
                  G_realloc ((char *) q->table, 
			  q->maxNofRules * sizeof (struct Quant_table));
  }
}

/*--------------------------------------------------------------------------*/

void
G_quant_set_neg_infinite_rule (q, dLeft, c)

     struct Quant *q;
     DCELL dLeft;
     CELL c;

{
  q->infiniteDLeft = dLeft;
  q->infiniteCLeft = c;
  quant_update_limits (q, dLeft, dLeft, c, c); 
  q->infiniteLeftSet = 1;
}

/*--------------------------------------------------------------------------*/

int
G_quant_get_neg_infinite_rule (q, dLeft, c)

     struct Quant *q;
     DCELL *dLeft;
     CELL *c;

{
  if (q->infiniteLeftSet == 0) return 0;

  *dLeft = q->infiniteDLeft;
  *c = q->infiniteCLeft;

  return 1;
}

/*--------------------------------------------------------------------------*/

void
G_quant_set_pos_infinite_rule (q, dRight, c)

     struct Quant *q;
     DCELL dRight;
     CELL c;

{
  q->infiniteDRight = dRight;
  q->infiniteCRight = c;
  quant_update_limits (q, dRight, dRight, c, c); 
  q->infiniteRightSet = 1;
}

/*--------------------------------------------------------------------------*/

int
G_quant_get_pos_infinite_rule (q, dRight, c)

     struct Quant *q;
     DCELL *dRight;
     CELL *c;

{
  if (q->infiniteRightSet == 0) return 0;

  *dRight = q->infiniteDRight;
  *c = q->infiniteCRight;

  return 1;
}

/*--------------------------------------------------------------------------*/

void
G_quant_add_rule (q, dLow, dHigh, cLow, cHigh)

     struct Quant *q;
     DCELL dLow, dHigh;
     CELL cLow, cHigh;

{
  int i;
  struct Quant_table *p;

  quant_table_increase (q);

  i = q->nofRules;

  p = &(q->table[i]);
  if (dHigh >= dLow) {
    p->dLow = dLow; p->dHigh = dHigh; p->cLow = cLow; p->cHigh = cHigh;  
  } else {
    p->dLow = dHigh; p->dHigh = dLow; p->cLow = cHigh; p->cHigh = cLow;  
  }

  quant_update_limits (q, dLow, dHigh, cLow, cHigh);

  q->nofRules++;
}

/*--------------------------------------------------------------------------*/

void
G_quant_reverse_rule_order (q)

     struct Quant *q;

{
  struct Quant_table tmp;
  struct Quant_table *pLeft, *pRight;

  pLeft = q->table;
  pRight = &(q->table [q->nofRules - 1]);

  while (pLeft < pRight) {
    tmp.dLow = pLeft->dLow; tmp.dHigh = pLeft->dHigh;
    tmp.cLow = pLeft->cLow; tmp.cHigh = pLeft->cHigh;

    pLeft->dLow = pRight->dLow; pLeft->dHigh = pRight->dHigh;
    pLeft->cLow = pRight->cLow; pLeft->cHigh = pRight->cHigh;

    pRight->dLow = tmp.dLow; pRight->dHigh = tmp.dHigh;
    pRight->cLow = tmp.cLow; pRight->cHigh = tmp.cHigh;

    pLeft++; pRight--;
  }    
}

/*--------------------------------------------------------------------------*/

static CELL
quant_interpolate (dLow, dHigh, cLow, cHigh, dValue)

     DCELL dLow, dHigh;
     CELL cLow, cHigh;
     DCELL dValue;

{
  if (cLow == cHigh) return cLow;
  if (dLow == dHigh) return cLow;

  return (CELL) ((dValue - dLow) / (dHigh - dLow) * (DCELL) (cHigh - cLow) + 
		 (DCELL) cLow); 
}

/*--------------------------------------------------------------------------*/

CELL
G_quant_get_cell_value (q, cellVal)

     struct Quant *q;
     DCELL cellVal;

{
  CELL tmp;
  struct Quant_table *p;

  if (NO_EXPLICIT_RULE) return NO_DATA;

  if (! NO_FINITE_RULE)
    for (p = &(q->table [q->nofRules - 1]); p >= q->table; p--)
      if ((cellVal >= p->dLow) && (cellVal <= p->dHigh)) 
	return quant_interpolate (p->dLow, p->dHigh, p->cLow, p->cHigh,
				  cellVal);
  
  if ((! NO_LEFT_INFINITE_RULE) && (cellVal <= q->infiniteDLeft))
    return q->infiniteCLeft;

  if ((NO_RIGHT_INFINITE_RULE) || (cellVal < q->infiniteDRight)) 
    return NO_DATA;

  return q->infiniteCRight;
}

/*--------------------------------------------------------------------------*/

void
G_quant_perform_d (q, dcell, cell, n)

     struct Quant *q;
     DCELL *dcell;
     CELL *cell;
     int n;

{
  int i;

  for (i = 0; i < n; i++, dcell++)
    if (! G_is_d_null_value (dcell))
      *cell++ = G_quant_get_cell_value (q, *dcell);
    else
      G_set_c_null_value (cell++, 1);
}

/*--------------------------------------------------------------------------*/

void
G_quant_perform_f (q, fcell, cell, n)

     struct Quant *q;
     FCELL *fcell;
     CELL *cell;
     int n;

{
  int i;

  for (i = 0; i < n; i++, fcell++)
    if (! G_is_f_null_value (fcell))
      *cell++ = G_quant_get_cell_value (q, (DCELL) *fcell);
    else
      G_set_c_null_value (cell++, 1);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

