#include "gis.h"
#include <string.h>

/*********************************************************************
*
*   G_quantize_fp_map(name, min, max)
*   char *name;            name of the map
*   CELL min, max;         resulting int range
*
*   Writes necessary quant rules for map <name> so that
*   a floating range of <name> is mapped into integer range (min, max)
*
**********************************************************************
* 
*   G_quantize_fp_map_range(name, d_min, d_max, min, max)
*   char *name;            name of the map
*   CELL min, max;         resulting int range
*   DCELL d_min, d_max;    floating point range
*
*   Make a rule for map <name> that maps floating range (d_min, d_max)
*   into integer range (min, max)
*   This function is useful when the quant rule doesn't depend of the
*   range of produced float data, for example the slope map whould
*   want to have a quant rule: 0.0, 90.0 -> 0 , 90
*   no matter what the min and max slope of this map is.
*
**********************************************************************
* 
*   G_write_quant(name, quant)
*        char *name;
*        struct Quant *quant;
*   writes the quant rule table for the map <name> and changes all
*   the necessary support files for <name>
*
**********************************************************************
* 
*   G_read_quant(name, mapset, quant)
*        char *name;
*        struct Quant *quant;
*   reads the quant table for name@mapset
*
**********************************************************************/

G_quantize_fp_map(name, min, max)
    char *name;
    CELL min, max;
{
    char buf[300];
    DCELL d_min, d_max;
    struct FPRange fp_range;

    if(G_read_fp_range(name, G_mapset(), &fp_range) < 0)
    {
        sprintf(buf, "G_quantize_fp_map: can't read fp range for map %s", name);
        G_warning(buf);
        return -1;
    }
    G_get_fp_range_min_max(&fp_range, &d_min,  &d_max);
    if(G_is_d_null_value(&d_min) || G_is_d_null_value(&d_max))
    {
	 sprintf(buf, "G_quantize_fp_map: raster map %s is empty", name);
	 G_warning(buf);
	 return -1;
    }
    return G_quantize_fp_map_range(name, d_min, d_max, min, max);
}

/*-------------------------------------------------------------------------*/

G_quantize_fp_map_range(name, d_min, d_max, min, max)
    char *name;
    CELL min, max;
    DCELL d_min, d_max;
{
    char buf[300];
    struct Quant quant;

    G_quant_init(&quant);
    G_quant_add_rule(&quant, d_min, d_max, min, max);
    /* quantize the map */
    if(G_write_quant (name, &quant) < 0)
    {
        sprintf(buf, "G_quantize_fp_map_range: can't write quant rules for map %s", name);        G_warning(buf);
        return -1;
    }
    return 1;
}

/*-------------------------------------------------------------------------*/


int 
G_write_quant(name, quant)
     char *name;
     struct Quant *quant;
{
     struct Categories cats;
     struct Range range;
     CELL cell_min, cell_max;
     DCELL d_min, d_max;
     char buf[300];

     if (G_raster_data_type (name, G_mapset()) == CELL_TYPE)
     {
            sprintf(buf, "Cannot write quant rules: map %s is integer", name);
            G_warning(buf);
            return -1;
     }

     G_quant_get_limits (quant, &d_min, &d_max, &cell_min, &cell_max); 

     /* first actually write the rules */
     if( G__quant_export (name, quant) < 0)
     {
            sprintf(buf, "Cannot write quant rules for map %s", name);
            G_warning(buf);
            return -1;
     }

     /* write new category files */
     G_init_cats (cell_max, (char *)NULL, &cats);
     G_write_cats (name, &cats);
     G_free_cats (&cats);


     /* write new integer range file */
     G_init_range(&range);
     G_update_range(cell_min, &range);
     G_update_range(cell_max, &range);
     G_write_range (name, &range);
   
     return 1;
}

/*-------------------------------------------------------------------------*/

int
G_read_quant(name, mapset, quant)
     char *name, *mapset;
     struct Quant *quant;
{
     G_quant_init (quant);
     return G__quant_import(name, mapset, quant);
}
